/*
 * This file is part of the Owl Positioning System (OwlPS) project.
 * It is subject to the copyright notice and license terms in the
 * COPYRIGHT.t2t file found in the top-level directory of this
 * distribution and at
 *   http://code.lm7.fr/p/owlps/source/tree/master/COPYRIGHT.t2t
 * No part of the OwlPS Project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms
 * contained in the COPYRIGHT.t2t file; the COPYRIGHT.t2t file must be
 * distributed along with this file, either separately or by replacing
 * this notice by the COPYRIGHT.t2t file's contents.
 */


#ifndef _OWLPS_POSITIONING_CALIBRATIONREQUEST_HH_
#define _OWLPS_POSITIONING_CALIBRATIONREQUEST_HH_

#include "request.hh"
#include "direction.hh"
#include "referencepoint.hh"
#include "posutil.hh"

/// Represents a calibration Request sent by a mobile
class CalibrationRequest: public Request
{
protected:
  /// Reference point that the mobile calibrates
  ReferencePoint *reference_point ;
  /// Direction in which the mobile was during the measurement
  Direction direction ;

public:
  CalibrationRequest(const uint_fast8_t _type = OWL_REQUEST_AUTOCALIBRATION) ;

  CalibrationRequest(const CalibrationRequest &source):
    Request(source), reference_point(source.reference_point),
    direction(source.direction) {}

  CalibrationRequest(const Request &source,
                     ReferencePoint *const _reference_point = nullptr,
                     const Direction &_direction = Direction(),
                     const uint_fast8_t _type = OWL_REQUEST_AUTOCALIBRATION) ;

  ~CalibrationRequest(void) {}

  /** @name Read accessors */
  //@{
  const Direction& get_direction(void) const ;
  ReferencePoint* get_reference_point(void) const ;
  //@}

  /** @name Write accessors */
  //@{
  void set_direction(const Direction &_direction) ;
  void set_reference_point(ReferencePoint *const _rp) ;
  /// Adds the CalibrationRequest to the #reference_point list of requests
  void reference_point_backward_link(void) const ;
  /// Deletes all the requests of #reference_point
  void reference_point_delete_requests(void) const ;
  void clear(void) ;
  //@}

  /** @name Operators */
  //@{
  CalibrationRequest& operator=(const CalibrationRequest &source) ;
  bool operator==(const CalibrationRequest &source) const ;
  bool operator!=(const CalibrationRequest &source) const ;
  //@}
} ;



/* *** Read accessors *** */


inline ReferencePoint* CalibrationRequest::get_reference_point(void) const
{
  return reference_point ;
}


inline const Direction& CalibrationRequest::get_direction(void) const
{
  return direction ;
}



/* *** Write accessors *** */


inline void CalibrationRequest::
set_reference_point(ReferencePoint *const _rp)
{
  reference_point = _rp ;
}


inline void CalibrationRequest::
set_direction(const Direction &_direction)
{
  direction = _direction ;
}



/* *** Operators *** */


inline bool CalibrationRequest::
operator!=(const CalibrationRequest &source) const
{
  return !(*this == source) ;
}



namespace std
{
template<> struct hash<CalibrationRequest>
{
public:
  size_t operator()(const CalibrationRequest &source) const
  {
    size_t seed = 0 ;

    PosUtil::hash_combine(seed, static_cast<Request>(source)) ;
    PosUtil::hash_combine(seed, source.get_direction()) ;
    if (source.get_reference_point())
      PosUtil::hash_combine(seed, *source.get_reference_point()) ;

    return seed ;
  }
} ;
}



#endif // _OWLPS_POSITIONING_CALIBRATIONREQUEST_HH_
