/*
 * This file is part of the Owl Positioning System (OwlPS) project.
 * It is subject to the copyright notice and license terms in the
 * COPYRIGHT.t2t file found in the top-level directory of this
 * distribution and at
 *   http://code.lm7.fr/p/owlps/source/tree/master/COPYRIGHT.t2t
 * No part of the OwlPS Project, including this file, may be copied,
 * modified, propagated, or distributed except according to the terms
 * contained in the COPYRIGHT.t2t file; the COPYRIGHT.t2t file must be
 * distributed along with this file, either separately or by replacing
 * this notice by the COPYRIGHT.t2t file's contents.
 */


#ifndef _OWLPS_POSITIONING_WIFIDEVICE_HH_
#define _OWLPS_POSITIONING_WIFIDEVICE_HH_

#include "posutil.hh"

#include <string>
#include <ostream>

#define WIFIDEVICE_DEFAULT_ANTENNA_GAIN 2
#define WIFIDEVICE_DEFAULT_TRX_POWER 20 // 20 dBm = 100 mW

/// Represents a Wi-Fi-enabled device
/**
 * This class is virtual by design, you should derive it to create new
 * types of Wi-Fi devices.
 */
class WifiDevice
{
protected:
  std::string ip_addr ;
  std::string mac_addr ;
  float antenna_gain ; ///< Antenna gain in dBi
  float trx_power ; ///< Transmit power in dBm

public:
  WifiDevice(const std::string &_ip_addr = "",
             const std::string &_mac_addr = "",
             const float _antenna_gain = WIFIDEVICE_DEFAULT_ANTENNA_GAIN,
             const float _trx_power = WIFIDEVICE_DEFAULT_TRX_POWER) ;

  WifiDevice(const WifiDevice &source):
    ip_addr(source.ip_addr), mac_addr(source.mac_addr),
    antenna_gain(source.antenna_gain), trx_power(source.trx_power) {}

  virtual ~WifiDevice(void) {}

  /** @name Read accessors */
  //@{
  const std::string& get_ip_addr(void) const ;
  const std::string& get_mac_addr(void) const ;
  float get_antenna_gain(void) const ;
  float get_trx_power(void) const ;
  //@}

  /** @name Write accessors */
  //@{
  void set_ip_addr(const std::string &_ip_addr) ;
  void set_mac_addr(const std::string &_mac_addr) ;
  void set_antenna_gain(const float _antenna_gain) ;
  void set_trx_power(const float _trx_power) ;
  //@}

  /** @name Operators */
  //@{
  WifiDevice& operator=(const WifiDevice &wd) ;
  bool operator==(const WifiDevice &wd) const ;
  bool operator!=(const WifiDevice &wl) const ;
  //@}

  /// Displays a WifiDevice
  friend std::ostream &operator<<(std::ostream &os, const WifiDevice &wd) ;
} ;



/* *** Constructors *** */


/**
 * `_mac_addr` must be uppercase.
 */
inline WifiDevice::
WifiDevice(const std::string &_ip_addr,
           const std::string &_mac_addr,
           const float _antenna_gain,
           const float _trx_power):
  ip_addr(_ip_addr), mac_addr(_mac_addr),
  antenna_gain(_antenna_gain), trx_power(_trx_power)
{
  assert_uppercase(mac_addr) ;
}



/* *** Read accessors *** */


inline const std::string& WifiDevice::get_ip_addr() const
{
  return ip_addr ;
}


inline const std::string& WifiDevice::get_mac_addr() const
{
  return mac_addr ;
}


inline float WifiDevice::get_antenna_gain() const
{
  return antenna_gain ;
}


inline float WifiDevice::get_trx_power() const
{
  return trx_power ;
}



/* *** Write accessors *** */


inline void WifiDevice::set_ip_addr(const std::string &_ip_addr)
{
  ip_addr = _ip_addr ;
}


inline void WifiDevice::set_mac_addr(const std::string &_mac_addr)
{
  assert_uppercase(_mac_addr) ;
  mac_addr = _mac_addr ;
}


inline void WifiDevice::set_antenna_gain(const float _antenna_gain)
{
  antenna_gain = _antenna_gain ;
}


inline void WifiDevice::set_trx_power(const float _trx_power)
{
  trx_power = _trx_power ;
}



/* *** Operators *** */


inline bool WifiDevice::operator!=(const WifiDevice &wd) const
{
  return !(*this == wd) ;
}



#endif // _OWLPS_POSITIONING_WIFIDEVICE_HH_
